@@ -38,7 +38,9 @@ from tqdm import tqdm
 # check if kerberos is installed
 try:
     from krb5._exceptions import Krb5Error
-    import gssapi
+    from gssapi.creds import Credentials as GSSAPICredentials
+    from gssapi.raw import Creds as RawCreds
+    from gssapi.exceptions import MissingCredentialsError, ExpiredCredentialsError
 
     is_kerb_available = True
 except ImportError:
@@ -60,6 +62,7 @@ MENU_COMMANDS = [
     "upload",
     "download",
     "loadps",
+    "runps",
     "menu",
     "clear",
     "exit",
@@ -110,7 +113,7 @@ class DelayedKeyboardInterrupt:
             self.old_handler(*self.signal_received)
 
 
-def run_ps(r_pool: RunspacePool, command: str) -> tuple[str, list, bool]:
+def run_ps_cmd(r_pool: RunspacePool, command: str) -> tuple[str, list, bool]:
     """Runs a PowerShell command and returns the output, streams, and error status."""
     log.info("Executing command: {}".format(command))
     ps = PowerShell(r_pool)
@@ -122,7 +125,7 @@ def run_ps(r_pool: RunspacePool, command: str) -> tuple[str, list, bool]:
 
 def get_prompt(r_pool: RunspacePool) -> str:
     """Returns the prompt string for the interactive shell."""
-    output, streams, had_errors = run_ps(
+    output, streams, had_errors = run_ps_cmd(
         r_pool, "$pwd.Path"
     )  # Get current working directory
     if not had_errors:
@@ -138,6 +141,7 @@ def show_menu() -> None:
         ("upload <local_path> <remote_path>", "Upload a file"),
         ("download <remote_path> <local_path>", "Download a file"),
         ("loadps <local_path>.ps1", "Load PowerShell functions from a local script"),
+        ("runps <local_path>.ps1", "Run a local PowerShell script on the remote host"),
         ("menu", "Show this menu"),
         ("clear, cls", "Clear the screen"),
         ("exit", "Exit the shell"),
@@ -182,7 +186,7 @@ def get_remote_path_suggestions(
     attrs = ""
     if not re.match(r"^[a-zA-Z]:", directory_prefix):
         # If the path doesn't start with a drive letter, prepend the current directory
-        pwd, streams, had_errors = run_ps(
+        pwd, streams, had_errors = run_ps_cmd(
             r_pool, "$pwd.Path"
         )  # Get current working directory
         directory_prefix = f"{pwd}\\{directory_prefix}"
@@ -220,17 +224,16 @@ def get_local_path_suggestions(
                         f"{entry}{os.sep}"  # Append a trailing slash for directories
                     )
                     if directory_prefix.startswith("~"):
-                        # If the directory prefix starts with ~, replace it with the home directory
-                        entry = str(entry).replace("~", home, 1)
-                    entry = str(entry).replace(home, "~", 1)
+                        # If the directory prefix starts with ~, replace home with ~
+                        entry = str(entry).replace(home, "~", 1)
                     suggestions.append(str(entry))
                 else:
                     if (extension is None) or (
                         entry.suffix.lower() == extension.lower()
                     ):
                         if directory_prefix.startswith("~"):
-                            # If the directory prefix starts with ~, replace it with the home directory
-                            entry = str(entry).replace("~", home, 1)
+                            # If the directory prefix starts with ~, replace home with ~
+                            entry = str(entry).replace(home, "~", 1)
                         suggestions.append(str(entry))
     except (FileNotFoundError, NotADirectoryError, PermissionError):
         pass
@@ -262,6 +265,31 @@ class CommandPathCompleter(Completer):
 
         command_typed_part = tokens[0]
 
+        # Handle .\name or ./name as first-token paths (run from current remote directory)
+        if command_typed_part.startswith(".\\") or command_typed_part.startswith("./"):
+            path_being_completed = command_typed_part
+            # strip surrounding quotes if any
+            if path_being_completed.startswith('"') and path_being_completed.endswith(
+                '"'
+            ):
+                path_being_completed = path_being_completed.strip('"')
+            directory_prefix, partial_name = get_directory_and_partial_name(
+                path_being_completed, sep="\\"
+            )
+            suggestions = get_remote_path_suggestions(
+                self.r_pool, directory_prefix, partial_name
+            )
+            for sugg_path in suggestions:
+                text_to_insert_in_prompt = f".\\" + sugg_path
+                if " " in sugg_path:
+                    text_to_insert_in_prompt = f'& ".\\{sugg_path}"'
+                yield Completion(
+                    text_to_insert_in_prompt,
+                    start_position=-len(command_typed_part),
+                    display=sugg_path,
+                )
+            return
+
         # Case 1: Completing the command name itself
         # There's only one token and no trailing space.
         if len(tokens) == 1 and not text_before_cursor.endswith(" "):
@@ -409,7 +437,7 @@ class CommandPathCompleter(Completer):
             else:
                 # More than 2 arguments, e.g., "download remote_path local_path extra_arg"
                 pass
-        elif actual_command_name == "loadps":
+        elif actual_command_name in ["loadps", "runps"]:
             # syntax: loadps <local_path>
             num_args_present = len(args)
 
@@ -750,11 +778,6 @@ def load_ps(r_pool: RunspacePool, local_path: str):
             # Find all function names in the script
             pattern = r"function\s+([a-zA-Z0-9_-]+)\s*(?={|$)"
             function_names = re.findall(pattern, content, re.MULTILINE)
-            global COMMAND_SUGGESTIONS
-            # Update the command suggestions with the function names
-            for func in function_names:
-                if func not in COMMAND_SUGGESTIONS:
-                    COMMAND_SUGGESTIONS += [func]
 
         ps.add_script(f". {{ {script} }}")  # Dot sourcing the script
         ps.begin_invoke()
@@ -774,6 +797,56 @@ def load_ps(r_pool: RunspacePool, local_path: str):
         else:
             print(GREEN + "[+] PowerShell script loaded successfully." + RESET)
             log.info(f"PowerShell script '{local_path}' loaded successfully.")
+            global COMMAND_SUGGESTIONS
+            # Update the command suggestions with the function names
+            new_suggestions = []
+            for func in function_names:
+                if func not in COMMAND_SUGGESTIONS:
+                    new_suggestions += [func]
+            if new_suggestions:
+                COMMAND_SUGGESTIONS += new_suggestions
+                print(
+                    CYAN
+                    + "[*] New commands available (use TAB to autocomplete):"
+                    + RESET
+                )
+                print(", ".join(new_suggestions))
+    except KeyboardInterrupt:
+        if ps.state == PSInvocationState.RUNNING:
+            log.info("Stopping command execution.")
+            ps.stop()
+
+
+def run_ps(r_pool: RunspacePool, local_path: str) -> None:
+    """Runs a local PowerShell script on the remote host."""
+    ps = PowerShell(r_pool)
+    try:
+        with open(local_path, "r") as script_file:
+            script = script_file.read()
+
+        ps.add_script(script)
+        ps.begin_invoke()
+
+        cursor = 0
+        while ps.state == PSInvocationState.RUNNING:
+            with DelayedKeyboardInterrupt():
+                ps.poll_invoke()
+            output = ps.output
+            for line in output[cursor:]:
+                print(line)
+            cursor = len(output)
+
+        if ps.streams.error:
+            print(RED + "[-] Failed to run PowerShell script." + RESET)
+            log.error(f"Failed to run PowerShell script '{local_path}'.")
+            for error in ps.streams.error:
+                print(RED + error._to_string + RESET)
+                log.error("Error: {}".format(error._to_string))
+                log.error("\tCategoryInfo: {}".format(error.message))
+                log.error("\tFullyQualifiedErrorId: {}".format(error.fq_error))
+        else:
+            print(GREEN + "[+] PowerShell script ran successfully." + RESET)
+            log.info(f"PowerShell script '{local_path}' ran successfully.")
     except KeyboardInterrupt:
         if ps.state == PSInvocationState.RUNNING:
             log.info("Stopping command execution.")
@@ -833,7 +906,7 @@ def interactive_shell(r_pool: RunspacePool) -> None:
                 remote_path = command_parts[1].strip('"')
                 local_path = command_parts[2].strip('"')
 
-                remote_file, streams, had_errors = run_ps(
+                remote_file, streams, had_errors = run_ps_cmd(
                     r_pool, f"(Resolve-Path -Path '{remote_path}').Path"
                 )
                 if not remote_file:
@@ -875,7 +948,7 @@ def interactive_shell(r_pool: RunspacePool) -> None:
 
                 if not re.match(r"^[a-zA-Z]:", remote_path):
                     # If the path doesn't start with a drive letter, prepend the current directory
-                    pwd, streams, had_errors = run_ps(r_pool, "$pwd.Path")
+                    pwd, streams, had_errors = run_ps_cmd(r_pool, "$pwd.Path")
                     if remote_path == ".":
                         remote_path = f"{pwd}\\{file_name}"
                     else:
@@ -884,7 +957,9 @@ def interactive_shell(r_pool: RunspacePool) -> None:
                 if remote_path.endswith("\\"):
                     remote_path = f"{remote_path}{file_name}"
 
-                upload_file(r_pool, str(Path(local_path).resolve()), remote_path)
+                upload_file(
+                    r_pool, str(Path(local_path).expanduser().resolve()), remote_path
+                )
                 continue
             elif command_lower.startswith("loadps"):
                 command_parts = quoted_command_split(command)
@@ -911,6 +986,31 @@ def interactive_shell(r_pool: RunspacePool) -> None:
 
                 load_ps(r_pool, local_path)
                 continue
+            elif command_lower.startswith("runps"):
+                command_parts = quoted_command_split(command)
+                if len(command_parts) < 2:
+                    print(RED + "[-] Usage: runps <local_path>" + RESET)
+                    continue
+                local_path = command_parts[1].strip('"')
+                local_path = Path(local_path).expanduser().resolve()
+
+                if not local_path.exists():
+                    print(
+                        RED
+                        + f"[-] Local PowerShell script '{local_path}' does not exist."
+                        + RESET
+                    )
+                    continue
+                elif local_path.suffix.lower() != ".ps1":
+                    print(
+                        RED
+                        + "[-] Please provide a valid PowerShell script file with .ps1 extension."
+                        + RESET
+                    )
+                    continue
+
+                run_ps(r_pool, local_path)
+                continue
             else:
                 try:
                     ps = PowerShell(r_pool)
@@ -975,21 +1075,6 @@ def main():
     parser.add_argument("-u", "--user", help="username")
     parser.add_argument("-p", "--password", help="password")
     parser.add_argument("-H", "--hash", help="nthash")
-    parser.add_argument(
-        "--no-pass", action="store_true", help="do not prompt for password"
-    )
-    if is_kerb_available:
-        parser.add_argument(
-            "-k", "--kerberos", action="store_true", help="use kerberos authentication"
-        )
-        parser.add_argument(
-            "--spn-prefix",
-            help="specify spn prefix",
-        )
-        parser.add_argument(
-            "--spn-hostname",
-            help="specify spn hostname",
-        )
     parser.add_argument(
         "--priv-key-pem",
         help="local path to private key PEM file",
@@ -1004,10 +1089,25 @@ def main():
         default="Microsoft WinRM Client",
         help='user agent for the WinRM client (default: "Microsoft WinRM Client")',
     )
-    parser.add_argument("--ssl", action="store_true", help="use ssl")
     parser.add_argument(
         "--port", type=int, default=5985, help="remote host port (default 5985)"
     )
+    if is_kerb_available:
+        parser.add_argument(
+            "--spn-prefix",
+            help="specify spn prefix",
+        )
+        parser.add_argument(
+            "--spn-hostname",
+            help="specify spn hostname",
+        )
+        parser.add_argument(
+            "-k", "--kerberos", action="store_true", help="use kerberos authentication"
+        )
+    parser.add_argument(
+        "--no-pass", action="store_true", help="do not prompt for password"
+    )
+    parser.add_argument("--ssl", action="store_true", help="use ssl")
     parser.add_argument("--log", action="store_true", help="log session to file")
     parser.add_argument("--debug", action="store_true", help="enable debug logging")
     parser.add_argument("--no-colors", action="store_true", help="disable colors")
@@ -1102,15 +1202,21 @@ def main():
                 )  # can also be cifs, ldap, HOST
                 if not args.user:
                     try:
-                        cred = gssapi.creds.Credentials(gssapi.raw.Creds())
+                        cred = GSSAPICredentials(RawCreds())
                         username = cred.name
-                    except gssapi.raw.exceptions.MissingCredentialsError:
+                    except MissingCredentialsError:
                         print(
                             MAGENTA
                             + "[%] No credentials cache found for Kerberos authentication."
                             + RESET
                         )
                         sys.exit(1)
+                    except ExpiredCredentialsError as ece:
+                        print(
+                            RED + "[-] The Kerberos credentials have expired. " + RESET
+                        )
+                        log.error("Expired credentials error: {}".format(ece))
+                        sys.exit(1)
                 # User needs to set environment variables `KRB5CCNAME` and `KRB5_CONFIG` as per requirements
                 # example: export KRB5CCNAME=/tmp/krb5cc_1000
                 # example: export KRB5_CONFIG=/etc/krb5.conf
